* Program BlockedMetabs.gms
********************************************************************************
$eolcom !

********************************************************************************
*Set Run Time
$if not set timeLimit $set timeLimit 20000

****************************
*Print Out FBA Options File*
****************************
$onecho > cplex.opt
eprhs 1e-9
epopt 1e-9
epint 0
tilim %timeLimit%
$offecho

********************************************************************************
******************************************
*Read in Stoichiometric Matrix and Limits*
******************************************
$include GPR_Toy.gms
$include Limits.gms
*$include Conditions.gms

********************************************************************************
************************
* Compose Minimal Media*
************************
Parameters
ModUpperLimits(j)
ModLowerLimits(j);

ModUpperLimits(j) = UpperLimits(j);
ModLowerLimits(j) = LowerLimits(j);

*Eliminate Metabolite Exchanges that are not in MediaSet or Essential
*ModLowerLimits(exch)$(not Essential(exch)) = 0;
*ModLowerLimits(exch)$MediaSet(exch,'1') = -10;

Set
biorxn(j) /vbio/;

********************************************************************************
***************************
* Set Solver Settings Here*
***************************
Parameters
iterationLimit   number of iterations allowed to solve problem /1e8/
relativeGap      relative gap used when solving MIP problem /0/;

********************************************************************************
********************************
* Initialize Relevant Variables*
********************************
Parameter
v_min minimum acceptable production levels for biomass /0.0005/
killW(i)
;

killW(i) = 0;

Binary Variables
w(i)
;

Free Variables
Obj
v(j)
;

Positive Variables
v_comp(i)
;
********************************************************************************
******************
* Formulate Model*
******************
Equations
Calc_Obj

*Primal Equations
MassBalance(i)
W_Indic(i)
UpperLimitCst(j)
LowerLimitCst(j)
killWCst(i)
;

Calc_Obj..               Obj =e= sum(i,w(i));

*Primal
MassBalance(i)..          sum(j,S(i,j)*v(j)) - v_comp(i) =e= 0;
UpperLimitCst(j)..        v(j) =l= ModUpperLimits(j);
LowerLimitCst(j)..        -v(j) =l= -ModLowerLimits(j);
W_Indic(i)..              v_comp(i) =g= w(i)*v_min;
killWCst(i)$(killW(i))..  w(i) =e= 0;

********************************************************************************
Model BlockedMetabIP /all/;
BlockedMetabIP.OptFile = 1;              !Default optFile to use
BlockedMetabIP.iterlim = iterationLimit; !Iteration Limit
BlockedMetabIP.optcr = relativeGap;      !relative gap

********************************************************************************
Obj.l = 1;
while(Obj.l > 0,
         SOLVE BlockedMetabIP maximizing Obj using mip;
         killW(i)$(killW(i) = 0) = w.l(i);
);

Parameters
cardi
wCount;

cardi = card(i);
wCount = sum(i,killW(i));

display cardi,wcount,killW;

Set
BlockedMetabs(i)
;

BlockedMetabs(i)$(not killW(i)) = yes;
*Output to .gdx
execute_unload "BlockMetabs.gdx" BlockedMetabs;

********************************************************************************
*************************
* Blocked Reaction List *
*************************
Parameter
BlockedRxn(j)
;

BlockedRxn(j) = 1;

Binary Variables
a(j)
Reverse(j)
;

Positive Variables
v_for(j)
v_rev(j)
;

Equations
Calc_Obj_Rxn

*Primal Equations
MassBalanceRxn(i)
A_Indic(j)
UpperLimitCstRxn(j)
LowerLimitCstRxn(j)
BlockedRxnCst(j)
;

Calc_Obj_Rxn..                             Obj =e= sum(j,a(j));

*Primal
MassBalanceRxn(i)..                        sum(j,S(i,j)*v_for(j)-S(i,j)*v_rev(j)) - v_comp(i) =e= 0;
UpperLimitCstRxn(j)..                      v_for(j) =l= ModUpperLimits(j)*(1-Reverse(j));
LowerLimitCstRxn(j)..                      v_rev(j) =l= -ModLowerLimits(j)*Reverse(j);
A_Indic(j)$(not biorxn(j))..               v_for(j) + v_rev(j) =g= a(j)*v_min;
BlockedRxnCst(j)$(not BlockedRxn(j))..     a(j) =e= 0;

********************************************************************************
Model BlockedRxnIP /Calc_Obj_Rxn,MassBalanceRxn,UpperLimitCstRxn,LowerLimitCstRxn,A_Indic,BlockedRxnCst/;
BlockedRxnIP.OptFile = 1;              !Default optFile to use
BlockedRxnIP.iterlim = iterationLimit; !Iteration Limit
BlockedRxnIP.optcr = relativeGap;      !relative gap

********************************************************************************
Obj.l = 1;
while(Obj.l > 0,
         SOLVE BlockedRxnIP maximizing Obj using mip;
         BlockedRxn(j)$(BlockedRxn(j) = 1) = 1-a.l(j);
);

Parameters
cardj
cardBlocked
blockedRxnCount
;

cardj = card(j);

blockedRxnCount = sum(j,BlockedRxn(j));

display cardj,BlockedRxnCount,BlockedRxn;

*Output to .gdx
execute_unload "BlockRxn.gdx" BlockedRxn;
