********************************************************************************
*GenExpMetab.gms
********************************************************************************

$eolcom !

$if not set EXCLUDE $set EXCLUDE vbio
*Set CPLEX Options
*Time Limit [s]
$if not set timeLimit $set timeLimit 5400
*Prune Time Limit [s]
$if not set pruneTimeLimit $set pruneTimeLimit 600
*Time to Continue Search after first couple found [s]
$if not set postCoupleLimit $set postCoupleLimit 600
Parameters
iterationLimit   number of iterations allowed to solve problem /50000/
relativeGap      relative gap used when solving MIP problem /0/;

$include GPR_Toy.gms
$include Limits.gms
$include User_Options.gms
********************************************************************************
********************************
*Generate Blocked Reaction List*
********************************
Sets
dualblocked(j)
$onempty
ManualBlock(g) used to manually set gene as blocked (fixes ko to 1) //
$offempty;

*Load GDX File Obtained from Program BlockMaker
execute_load "BlockRxn.gdx", dualblocked=BlockedRxn;

*Map genes to their reactions
Set
geneMap(g,n)
blockedGenes(g);

*Initialize Sets
geneMap(g,n)=no;
blockedGenes(g)=yes;

*Determine blocked genes
*Map gene to reaction if a rule exists and that reaction is not blocked
geneMap(g,n)$sum(j$(not dualblocked(j) and Rules_Reaction(j,n,g)),1)=yes;
*A gene is blocked only if it is not mapped to any unblocked reactions
blockedGenes(g)$sum(n$geneMap(g,n),1)=no;

blockedGenes(g)$ManualBlock(g)=yes;

display blockedGenes,dualblocked

********************************************************************************
************************************************
*Set-up Primal and Dual Problems for Evaluation*
************************************************
Parameters
*USER DEFINED PARAMTERS
maxKnockouts     maximum number of knockouts
minKnockouts     minimum number of knockouts
maxCarbon        maximum number of carbon sources
maxNitrogen      maximum number of nitrogen sources
maxAcceptors     maximum number of acceptors
maxAuxotrophs    maximum number of auxotroph metabolites
deletionPenalty  size of Deletion penalty (0 <= penalty <= 1)
auxotrophPenalty size of Auxotroph penalty (0 <= penalty < 1.0)
iterationLimit   number of iteations allowed to solve problem
relativeGap      relative gap used when solving MIP problem
epsilon          Condition for acceptance of coupling;

$if not set EXCLUDE $abort EXCLUDE is not defined
Sets
*USER DEFINED SETS
$onempty
biorxn(j)         Biomass Reaction Flux /%EXCLUDE%/
biometabList(i)   List of metabs to be removed /C,E/
biometab(i)       metab to be removed //
Carbon(j)         subset of carbon source media available
Nitrogen(j)       subset of nitrogen sources available
Acceptors(j)      subset of electron acceptors available
Auxotrophs(j)     subset of metabolites for auxotrophic scenarios
Essential(j)      set of set of essential items in minimal media
EssentialGenes(g) set of essential genes
Deleter(g)        set of user mandated gene deletions //
$offempty
*POSTPROCESSED SETS (Use Method SetConditions)
remainder(j)     Set of unused exchanges
aset(j)          Set over which deleter traverses
notaset(j)       Set untouched by deleter
Overlap(j)       Master Set for Media Deletions (handles overlaps)
fixGenes(g)      Genes to be fixed because they are mapped\blocked\essential
ErrorTest(j)     Used to make certain there is no intersection with overlap and essential
ErrorTest2(j)    Used to make certain that Overlap and Essential have lowerlimits
ErrorTest3(j)    Used to make certain that Overlap and Essential are not blocked;

*Define area over which deletions are allowed
notaset(j)=jl(j);
aset(j)=(not notaset(j));

*MIP Variables
Binary Variable
a(j)     MIP variable used to delete reactions
b(j)     MIP variable used to select carbon source
c(j)     MIP variable select Acceptors conditions
e(j)     MIP variable used to select Nitrogen source
f(j)     MIP variable used to select Auxotroph Metabolite
h(j)     MIP variable used to handle overlapping variables
ko(g)    MIP variable used to delete genes;

*Primal Problem
Positive Variables
v_sink(i)        sink variable used to determine if compound is produced
holder(j,n)      holder for protein rules;
holder.up(j,n)=1;

Free Variable
v(j)             flux variables
;

*Dual Problem
Positive Variables
u_upper(j)          u associated with upper bound constraints
u_lower(j)          u associated with lower bound constraints
umedia(j)           u associated with t constraint
ubio                u associated with t constraint;

Free Variables
u_mat(i)         u associated with matrix coefficients
udel(j)          u associated with deleted reactions
;

*Outer Objective
Free Variables
Outer_Obj        Secondary Objective of Bilinear Optimization;
********************************************************************************
**************************************
*Declare Bilinear Problem Constraints*
**************************************

Equations
*Primal Problem Equations
massbalance(i)         mass balance equations for each metabolite
UpperLimitCst(j)       Upper Bound Flux Constraints
LowerLimitCst(j)       Lower Bound Flux Constraints

*Primal Deletion Constraints
Rxn_killer(j)            Kills reactions that have been deleted from the network
Exchange_Killer(j)       Kills reverse exchange flux when overlap is not activated
Remainder_Killer(j)      Kills exchange fluxes that have not been specified to be in model
Bio_Cst(j)               Guarantees Strain Growth

*Dual Problem Equations
gen_flux_dual(j)         dual constraints for quintessential forward fluxes
sink_dual(i)             dual constraint associated with variable v_sink

*Dual Deletion Constraints
udelcst(j)
umediacst(j)

*Equations Related to Objective
Calc_PrimalDual_Obj      Calculate objective (PrimalDual)
Calc_Outer_Obj           Calculate primary objective

*GPR Rules
prules1(j,n)             determine protein states from gene deletions
prules2(j,n,g)           determine protein states from gene deletions
prules3(j)               determine reaction states from enzyme comples states
prules4(j,n)             determine reaction states from enzyme comples states

*Overlap Rules
Overlap_Rule_0(j)        b(j) + c(j) + e(j) + f(j) <= 1 (Do not use one metabolite for multiple sources)
Overlap_Rule_1(j)        h(j)=0 when all are false
Overlap_Rule_2(j)        h(j)=1 if b(j)=1
Overlap_Rule_3(j)        h(j)=1 if c(j)=1
Overlap_Rule_4(j)        h(j)=1 if e(j)=1
Overlap_Rule_5(j)        h(j)=1 if f(j)=1

*Constraints on MIP Variabls
maxgene                  Maximum number of genes to be deleted
mingene                  Minimum number of genes to be deleted
maxmedia                 Maximum number of carbon sources allowed
maxnitro                 Maximum number of nitrogen sources allowed
maxaccept                Maximum number of electron acceptors to be allowed
maxauxo                  Maximum number of autotrophic metabolites allowed
minaccept;

*****************************INNER OBJECTIVE************************************
*Primal Constraints
*max sum(biometab,v_sink(biometab))
massbalance(i)..                                 sum(j$(not dualblocked(j)), S(i,j)*v(j)) - v_sink(i)$bioMetab(i) =e= 0;
UpperLimitCst(j)$(not dualblocked(j))..          v(j) =l= UpperLimits(j);
LowerLimitCst(j)$(not dualblocked(j))..          -v(j) =l= -LowerLimits(j);

*Primal Deletion Constraints
Rxn_killer(aset)..                                v(aset)            =e= 0;
Exchange_Killer(Overlap)..                       -v(Overlap)         =l= 0;
Remainder_Killer(Remainder)..                    -v(Remainder)       =l= 0;
Bio_Cst(biorxn)..                                -v(biorxn)          =l= -epsilon;

*Dual Constraints
gen_flux_dual(j)$(not dualblocked(j))..          sum(i, S(i,j)*u_mat(i) ) + u_upper(j) - u_lower(j) + udel(j)$aset(j) - umedia(j)$(exch(j) and (not essential(j))) - ubio$biorxn(j) =e= 0;
sink_dual(biometab)..                            -u_mat(biometab) =g= 1;

*Dual Deletion Constraints
udelcst(aset)..                                  udel(aset)        =e= 0;
umediacst(Overlap)..                             umedia(Overlap)   =e= 0;

*Inner Objective Function
Calc_PrimalDual_Obj..                            sum(biometab,v_sink(biometab)) =e= sum(j$(not dualblocked(j)), UpperLimits(j)*u_upper(j) - LowerLimits(j)*u_lower(j) - epsilon*ubio$(biorxn(j)));

*****************************OUTER OBJECTIVE************************************
*Gene Deletion
prules1(rr(j,n)).. sum(g$Rules_Reaction(j,n,g), ko(g)-1)=l=holder(j,n)-1;
prules2(Rules_Reaction(j,n,g)).. ko(g)=g=holder(j,n);
prules3(aset).. a(aset)=l=sum(n$rr(aset,n), holder(aset,n));
prules4(rr(aset,n)).. a(aset)=g=holder(aset,n);

*Overlap Rules
Overlap_Rule_0(Overlap).. b(Overlap)$Carbon(Overlap) + c(Overlap)$Acceptors(Overlap) +
                          e(Overlap)$Nitrogen(Overlap) + f(Overlap)$Auxotrophs(Overlap) =l= 1;
Overlap_Rule_1(Overlap).. b(Overlap)$Carbon(Overlap) + c(Overlap)$Acceptors(Overlap) +
                          e(Overlap)$Nitrogen(Overlap) + f(Overlap)$Auxotrophs(Overlap) =g= h(Overlap);
Overlap_Rule_2(Overlap)$Carbon(Overlap)..        b(Overlap) =l= h(Overlap);
Overlap_Rule_3(Overlap)$Acceptors(Overlap)..     c(Overlap) =l= h(Overlap);
Overlap_Rule_4(Overlap)$Nitrogen(Overlap)..      e(Overlap) =l= h(Overlap);
Overlap_Rule_5(Overlap)$Auxotrophs(Overlap)..    f(Overlap) =l= h(Overlap);

*MIP Limitations
maxgene..                sum(g$(not Deleter(g)),1-ko(g))=l=maxKnockouts;
mingene..                sum(g,1-ko(g))=g=minKnockouts;
maxmedia..               sum(Carbon,b(Carbon)) =l= maxCarbon;
maxaccept..              sum(Acceptors,c(Acceptors)) =l= maxAcceptors;
minaccept..              sum(Acceptors,c(Acceptors)) =g= 1;
maxnitro..               sum(Nitrogen,e(Nitrogen)) =l= maxNitrogen;
maxauxo..                sum(Auxotrophs,f(Auxotrophs)) =l= maxAuxotrophs;

Calc_Outer_Obj..         Outer_Obj =e= (100/epsilon)*sum(biometab,v_sink(biometab)) + sum(g$(not Deleter(g)),1 - ko(g))*deletionPenalty + sum(Auxotrophs,f(Auxotrophs))*auxotrophPenalty;

********************************************************************************
**********************************
*Declare Model and Set Attributes*
**********************************

Model GenExp /all/;
GenExp.OptFile=1;              !Default optFile to use
GenExp.holdfixed=1;            !refrain from importing held vars
GenExp.iterlim=iterationLimit; !Iteration Limit
GenExp.optcr=relativeGap;      !relative gap

********************************************************************************
*Method Used To Process User Provided Data
$onecho > SetConditions.gms

*Create a Master Set of Source Exchanges
Overlap(j)=Carbon(j)+Nitrogen(j)+Acceptors(j)+Auxotrophs(j);

**************ERROR TESTING******************
*If the intersection of Overlap and Essential is not empty, throw error
ErrorTest(j)=Overlap(j)*Essential(j);
loop(ErrorTest,
         display ErrorTest;
         abort "Essential item placed in selective media list";
);
*If either essential or overlap do not have lowerlimits throw error
ErrorTest2(j)=Overlap(j)+Essential(j);
loop(j$(LowerLimits(j)),
         ErrorTest2(j)=no;
);
loop(ErrorTest2,
         display ErrorTest2;
         abort "Searched item has no Lower Limit";
);
*If either essential or overlap are blocked throw error
ErrorTest3(j) = Overlap(j)+Essential(j);
ErrorTest3(j) = ErrorTest3(j)*(dualblocked(j));
loop(ErrorTest3,
         display ErrorTest3;
         abort "Exchange reaction lives in blocked list";
);
if(auxotrophPenalty <0 or auxotrophPenalty >=1.0,
  abort "Auxotroph Penalty is outside allowed range. Present value:", auxotrophPenalty;
);

if(deletionPenalty <0 or deletionPenalty >1.0,
  abort "Deletion Penalty is outside allowed range. Present value:", deletionPenalty;
);
**********************************************
*Create a set containing all unused exchanges
remainder(j)=exch(j)-Overlap(j)-Essential(j);

*Undo any previous alteration to binary variables
a.lo(j)=0;
a.up(j)=1;
b.lo(j)=0;
b.up(j)=1;
c.lo(j)=0;
c.up(j)=1;
e.lo(j)=0;
e.up(j)=1;
f.lo(j)=0;
f.up(j)=1;
h.lo(j)=0;
h.up(j)=1;
ko.lo(g)=0;
ko.up(g)=1;

fixGenes(g) = blockedGenes(g)+EssentialGenes(g)+Subunit(g);

*Fix Unused Variables to Reduce Overhead
a.fx(notaset)=1;
b.fx(j)$(not Carbon(j))          =1;
c.fx(j)$(not Acceptors(j))       =1;
e.fx(j)$(not Nitrogen(j))        =1;
f.fx(j)$(not Auxotrophs(j))      =1;
h.fx(j)$(not Overlap(j))         =1;
ko.fx(g)$(fixGenes(g))           =1;
$offecho
********************************************************************************
*CREATE CPLEX OPTIONS FILE
$onecho > cplex.opt
eprhs 1e-9
epopt 1e-9
epint 0
numericalemphasis 1
tilim %timeLimit%
indic Rxn_killer(aset)$a(aset) 0
indic Exchange_Killer(Overlap)$h(Overlap) 0
indic udelcst(aset)$a(aset) 1
indic umediacst(Overlap)$h(overlap) 1
$offecho
********************************************************************************
Parameters
geneKOs(g)
Media(j)
StoreBioVal(i);

loop(biorxn,
         StoreBioVal(i) = S(i,biorxn);
);

Carbon(j) = minimalCarbon(j);
Nitrogen(j) = minimalNitrogen(j);
Acceptors(j)= minimalAcceptors(j);
$include SetConditions.gms

file result /GenerateExperimentsBatch.txt/;
result.pw=500; result.pc=5; result.ps=130; result.ap=1;
loop(biometabList,
         biometab(i) = no;
         biometab(bioMetabList) = yes;
         S(biometab,biorxn) = 0;
         geneKOs(g) = 0;
         Media(j) = 0;
         solve GenExp minimizing Outer_Obj using mip;
         geneKOs(g) = (1-ko.l(g));
         ko.fx(g)$geneKOs(g) = 1;
         Media(Overlap) = h.l(Overlap);
         Media(Essential) = h.l(Essential);
         S(biometab,biorxn) = StoreBioVal(biometab);

*Write Results
         put result;
         put "Biomass Component:"
         loop(Biometab,
                  put Biometab.tl;
         );
         put / /;
         put "Obj:" /;
         put Outer_Obj.l /;
         put "Model Stat/Solve Stat" /;
                 put GenExp.ModelStat GenExp.SolveStat /;
         put "Genes" /;
         loop(g$(geneKOs(g)),
                 put g.tl;
         );
         put /;
         put "Media" /;
         loop(j$(Media(j)),
                 put j.tl;
         );
         put / /;
         putclose result;
);
